
#include "disasm_shader.h"
#include "tools.h"
#include <map>
#include <cassert>

#define ASSERT_SRC() { if ( src == shader_code.end() ) throw ShaderError( "Unexpected end of shader", line_number ); }

namespace {

    REG_TYPE RegType( unsigned reg_type, unsigned reg_index )
    {
        switch ( REG_TYPE( reg_type ) ) {
        case REG_TEMP:      return REG_TYPE( REG_R0 + reg_index );
        case REG_INPUT:     return REG_TYPE( REG_V0 + reg_index );
        case REG_CONST:     return REG_TYPE( REG_C0 + reg_index );
        case REG_TEXTURE:   return REG_TYPE( REG_T0 + reg_index );
        }
        return REG_R0;
    }

}

void DisasmShader( const std::vector<unsigned>& shader_code, std::vector<INSTRUCTION>* p_instructions )
{
    assert( p_instructions );
    p_instructions->clear();
    std::vector<unsigned>::const_iterator src = shader_code.begin();
    int line_number = -1;
    ASSERT_SRC();

    // Check shader version
    const unsigned version_code = *(src++);
    const unsigned version_minor = version_code & 0xF;
    if ( (version_code & 0xFFFFFFF0) != 0xFFFF0100 || version_minor > 3 )
        throw ShaderError( "Unsupported shader version", -1 );

    // Feed in subsequent tokens until the end of shader
    const unsigned end_token = 0x0000FFFF;
    enum STAGE { DEFINITION, TEXTURE, ARITHMETIC };
    STAGE stage = DEFINITION;
    WRITE_MASK prev_instr_mask = WRITE_RGBA;
    unsigned prev_tex_def = 0xFFFFFFFF;
    bool c_def[] = { false, false };
    while ( src != shader_code.end() && *src != end_token ) {
        const unsigned token = *(src++);

        // Comment token
        if ( (token & 0x8000FFFF) == 0x0000FFFE ) {
            const unsigned num_tokens = token >> 16;
            if ( num_tokens == 2 && *src == 0x454E494C ) { // "LINE" backwards
                ASSERT_SRC();
                src++;
                ASSERT_SRC();
                line_number = int( *(src++) );
            } else
                for ( unsigned i=0; i < num_tokens; i++ ) {
                    ASSERT_SRC();
                    src++;
                }
                continue;
        }

        // Instruction
        if ( (token & 0x80000000) != 0 )
            throw ShaderError( "Unexpected register", line_number );
        const unsigned opcode = token & 0x0000FFFF;
        if ( opcode > OP_MUL && opcode != OP_DP3 && opcode != OP_LRP && opcode != OP_TEX && opcode != OP_TEXBEM && opcode != OP_TEXBEML && opcode != OP_DEF )
            throw ShaderError( "Unsupported instruction", line_number );
        const bool b_paired = (token & 0x40000000) != 0;
        if ( token & 0xBFFF0000 )
            throw ShaderError( "Unsupported instruction flags", line_number );

        // nop
        if ( opcode == OP_NOP )
            continue;

        // Check instruction vs. shader stage
        if ( opcode == OP_DEF ) {
            if ( stage != DEFINITION )
                throw ShaderError( "Misplaced def instruction", line_number );
        } else {
            if ( stage != ARITHMETIC )
                stage = TEXTURE;
            if ( opcode == OP_TEX || opcode == OP_TEXBEM || opcode == OP_TEXBEML ) {
                if ( stage != TEXTURE )
                    throw ShaderError( "Misplaced texturing instruction", line_number );
            } else
                stage = ARITHMETIC;
        }

        // Extract destination register
        ASSERT_SRC();
        const unsigned dest_reg_word = *(src++);
        if ( (dest_reg_word & 0x80000000) == 0 )
            throw ShaderError( "Unexpected instruction", line_number );
        const unsigned reg_index = dest_reg_word & 0x7FF;
        const unsigned reg_type = ((dest_reg_word>>28)&7) | ((dest_reg_word>>8)&0x18);
        signed char scale = (dest_reg_word >> 24) & 0xF;
        if ( scale > 7 ) scale = (scale ^ 0xF) - 1;
        if ( scale > 2 || scale < -1 )
            throw ShaderError( "Unsupported scale factor", line_number );
        const bool b_saturate = (dest_reg_word & 0xFFEFFFFF) != 0;
        const unsigned write_mask = (dest_reg_word >> 13) & 0x77F;
        WRITE_MASK write_type = WRITE_RGBA;
        if ( write_mask == 0x38 )
            write_type = WRITE_RGB;
        else if ( write_mask == 0x40 )
            write_type = WRITE_A;
        else if ( write_mask != 0x78 )
            throw ShaderError( "Unsupported write mask", line_number );

        // Check pairing
        if ( ! b_paired )
            prev_instr_mask = write_type;
        else {
            if ( stage != ARITHMETIC ||
                 (write_type != WRITE_RGB && write_type != WRITE_A) ||
                 (write_type == WRITE_RGB && prev_instr_mask != WRITE_A) ||
                 (write_type == WRITE_A && prev_instr_mask != WRITE_RGB) )
                throw ShaderError( "Instruction pairing error", line_number );
            prev_instr_mask = WRITE_RGBA;
        }

        // Check register type against instruction
        if ( opcode == OP_DEF ) {
            if ( write_type != WRITE_RGBA || scale != 0 || reg_type != REG_CONST || reg_index > 1 )
                throw ShaderError( "Invalid register", line_number );
        } else if ( stage == TEXTURE ) {
            if ( write_type != WRITE_RGBA || scale != 0 || reg_type != REG_TEXTURE || reg_index > 7 )
                throw ShaderError( "Invalid register", line_number );
            if ( prev_tex_def < 0xFFFFFFFF ) {
                if ( reg_index <= prev_tex_def )
                    throw ShaderError( "Textures must be declared in ascending order", line_number );
            }
            prev_tex_def = reg_index;
        } else
            if ( reg_type != REG_TEMP || reg_index > 1 )
                throw ShaderError( "Invalid register", line_number );

        // Build instruction
        INSTRUCTION inst;
        ZeroFill( &inst );
        inst.opcode = SHADER_OPCODE( opcode );
        inst.dest_reg = RegType( reg_type, reg_index );
        inst.write_mask = write_type;
        inst.dest_scale = scale;
        inst.b_saturate = b_saturate;
        inst.b_pair = b_paired;
        inst.line_number = line_number;

        // def
        if ( opcode == OP_DEF ) {
            if ( c_def[ reg_index ] )
                throw ShaderError( "Constant redefinition", line_number );
            c_def[ reg_index ] = true;
            FOR_INDEX( i, 4 ) {
                ASSERT_SRC();
                const float f = FloatFromU32( *(src++) );
                const int k = int( f * 255.f );
                inst.constant[i] = static_cast<unsigned char>( (k<0)?0:((k>255)?255:k) );
            }
            p_instructions->push_back( inst );
            continue;
        }

        // Determine number of required source operands
        size_t num_src_args = 2;
        if ( opcode == OP_TEX )
            num_src_args = 0;
        if ( opcode == OP_TEXBEM || opcode == OP_TEXBEML || opcode == OP_MOV )
            num_src_args = 1;
        else if ( opcode == OP_MAD || opcode == OP_LRP )
            num_src_args = 3;

        // Extract subsequent arguments
        FOR_INDEX( i, num_src_args ) {

            // Get arg dword
            ASSERT_SRC();
            const unsigned src_reg_word = *(src++);

            // Extract fields and check their valitity
            const unsigned reg_index = src_reg_word & 0x7FF;
            const unsigned reg_type = ((src_reg_word>>28)&7) | ((src_reg_word>>8)&0x18);
            if ( (src_reg_word & 0xE000) != 0 )
                throw ShaderError( "Invalid source register flags", line_number );
            const unsigned swizzle = (src_reg_word >> 16) & 0xFF;
            if ( swizzle != 0xFF && swizzle != 0xE4 && swizzle != 0xAA )
                throw ShaderError( "Unsupported source register swizzle", line_number );
            if ( swizzle == 0xAA && write_type != WRITE_A )
                throw ShaderError( "Unsupported source register swizzle (.b swizzle is supported only for .a write mask)", line_number );
            const bool b_replicate_alpha = (swizzle == 0xFF) && (write_type != WRITE_A);
            const bool b_replicate_blue = (swizzle == 0xAA) && (write_type == WRITE_A);
            const unsigned modifier = (src_reg_word >> 24) & 0xF;
            if ( modifier > MOD_INVERT )
                throw ShaderError( "Unsupported source register modifier", line_number );
            if ( stage == TEXTURE && (b_replicate_alpha || modifier != MOD_NONE) )
                throw ShaderError( "Modifiers not allowed on source operands to texbem[l]", line_number );

            // Remember argument
            inst.src_reg[i].reg = RegType( reg_type, reg_index );
            inst.src_reg[i].b_replicate_alpha = b_replicate_alpha;
            inst.src_reg[i].b_replicate_blue = b_replicate_blue;
            inst.src_reg[i].modifier = REG_MODIFIER( modifier );
        }
        p_instructions->push_back( inst );
    }

    // Check if shader ended correctly
    if ( src == shader_code.end() )
        throw ShaderError( "Unexpected end of shader", line_number );
}

std::string GetReadableShader( const std::vector<INSTRUCTION>& instructions )
{
    std::map<SHADER_OPCODE,std::string> map_shader_opcode;
    map_shader_opcode[ OP_MOV ] = "mov";
    map_shader_opcode[ OP_ADD ] = "add";
    map_shader_opcode[ OP_SUB ] = "sub";
    map_shader_opcode[ OP_MAD ] = "mad";
    map_shader_opcode[ OP_MUL ] = "mul";
    map_shader_opcode[ OP_DP3 ] = "dp3";
    map_shader_opcode[ OP_LRP ] = "lrp";
    map_shader_opcode[ OP_TEX ] = "tex";
    map_shader_opcode[ OP_TEXBEM ] = "texbem";
    map_shader_opcode[ OP_TEXBEML ] = "texbeml";
    map_shader_opcode[ OP_DEF ] = "def";
    std::map<WRITE_MASK,std::string> map_write_mask;
    map_write_mask[ WRITE_RGBA ] = "";
    map_write_mask[ WRITE_RGB ] = ".rgb";
    map_write_mask[ WRITE_A ] = ".a";
    std::map<REG_TYPE,std::string> map_reg_type;
    map_reg_type[ REG_R0 ] = "r0";
    map_reg_type[ REG_R1 ] = "r1";
    map_reg_type[ REG_V0 ] = "v0";
    map_reg_type[ REG_V1 ] = "v1";
    map_reg_type[ REG_C0 ] = "c0";
    map_reg_type[ REG_C1 ] = "c1";
    map_reg_type[ REG_T0 ] = "t0";
    map_reg_type[ REG_T1 ] = "t1";
    map_reg_type[ REG_T2 ] = "t2";
    map_reg_type[ REG_T3 ] = "t3";
    map_reg_type[ REG_T4 ] = "t4";
    map_reg_type[ REG_T5 ] = "t5";
    map_reg_type[ REG_T6 ] = "t6";
    map_reg_type[ REG_T7 ] = "t7";

    std::string outs;
    FOR_INDEX( i, instructions.size() ) {
        const INSTRUCTION& ins = instructions[i];
        if ( ins.b_pair ) outs += "+ ";
        else outs += "  ";
        outs += map_shader_opcode[ins.opcode];
        if ( ins.dest_scale == 1 )
            outs += "_x2";
        else if ( ins.dest_scale == 2 )
            outs += "_x4";
        else if ( ins.dest_scale == -1 )
            outs += "_d2";
        outs += " ";
        outs += map_reg_type[ins.dest_reg];
        outs += map_write_mask[ins.write_mask];
        if ( ins.opcode == OP_DEF ) {
            FOR_INDEX( i, 4 )
                outs += ", " + Int2String( ins.constant[i] );
        } else
            FOR_INDEX( i, 3 )
            if ( ins.src_reg[i].reg != REG_UNUSED ) {
                outs += ", ";
                const unsigned char modifier = ins.src_reg[i].modifier;
                if ( modifier == MOD_INVERT ) outs += "1-";
                else if ( modifier == MOD_NEGATE || modifier == MOD_NEGATE_BIAS || modifier == MOD_NEGATE_BIAS_X2 ) outs += "-";
                outs += map_reg_type[REG_TYPE(ins.src_reg[i].reg)];
                if ( modifier == MOD_BIAS || modifier == MOD_NEGATE_BIAS ) outs += "_bias";
                else if ( modifier == MOD_BIAS_X2 || modifier == MOD_NEGATE_BIAS_X2 ) outs += "_bx2";
                if ( ins.src_reg[i].b_replicate_alpha )
                    outs += ".a";
                if ( ins.src_reg[i].b_replicate_blue )
                    outs += ".b";
            }
        outs += "\n";
    }
    return outs;
}

// vim:ts=4:sts=4:et
