
#ifndef __DISASM_SHADER_H_INCLUDED
#define __DISASM_SHADER_H_INCLUDED

#include <string>
#include <vector>

/// Exception thrown by shader interpreting and loading routines to indicate errors
class ShaderError {
    std::string error_msg;  ///< Error string
    int line_number;        ///< Line number in source or -1 if no line number available
public:
    ShaderError( const char* error_msg, int line_number ): error_msg( error_msg ), line_number( line_number ) { }
    /// Returns error string
    const std::string& ErrorMsg() const { return error_msg; }
    /// Returns line number in source or -1 if no line number available
    int LineNumber() const { return line_number; }
};

/// Shader instruction opcode
enum SHADER_OPCODE {
    OP_NOP = 0,
    OP_MOV = 1,
    OP_ADD = 2,
    OP_SUB = 3,
    OP_MAD = 4,
    OP_MUL = 5,
    OP_DP3 = 8,
    OP_LRP = 18,
    OP_TEX = 66,
    OP_TEXBEM = 67,
    OP_TEXBEML = 68,
    OP_DEF = 81
};

/// Register write mask
enum WRITE_MASK {
    WRITE_RGBA,
    WRITE_RGB,
    WRITE_A
};

/// Register
enum REG_TYPE {

    // Register types
    REG_TEMP    = 0,    // r0, r1
    REG_INPUT   = 1,    // v0, v1
    REG_CONST   = 2,    // c0, c1
    REG_TEXTURE = 3,    // t0, t1, t2, t3, t4, t5, t6, t7

    // Registers extracted from binary shader code
    REG_UNUSED = 0,     ///< No register (unused slot)
    REG_R0,             ///< Temporary register 0
    REG_R1,             ///< Temporary register 1
    REG_V0,             ///< Diffuse color register
    REG_V1,             ///< Specular color register
    REG_C0,             ///< Constant register 0
    REG_C1,             ///< Constant register 1
    REG_T0,             ///< Texture register 0
    REG_T1,             ///< Texture register 1
    REG_T2,             ///< Texture register 2
    REG_T3,             ///< Texture register 3
    REG_T4,             ///< Texture register 4
    REG_T5,             ///< Texture register 5
    REG_T6,             ///< Texture register 6
    REG_T7              ///< Texture register 7
};

/// Register modifiers
enum REG_MODIFIER {
    MOD_NONE            = 0,    ///< X ([0,1] -> [0,1])
    MOD_NEGATE          = 1,    ///< -X ([0,1] -> [0,-1])
    MOD_BIAS            = 2,    ///< X_bias ([0,1] -> [-0.5,0.5])
    MOD_NEGATE_BIAS     = 3,    ///< -X_bias ([0,1] -> [0.5,-0.5])
    MOD_BIAS_X2         = 4,    ///< X_bx2 ([0,1] -> [-1,1])
    MOD_NEGATE_BIAS_X2  = 5,    ///< -X_bx2 ([0,1] -> [1,-1])
    MOD_INVERT          = 6     ///< 1-X ([0,1] -> [1,0])
};

/// Source register description
struct SRCREG {
    REG_TYPE reg;               ///< Register
    bool b_replicate_alpha;     ///< Alpha replication
    bool b_replicate_blue;      ///< Blue component replication
    REG_MODIFIER modifier;      ///< Register modifier
};

/// Instruction description
struct INSTRUCTION {
    SHADER_OPCODE opcode;       ///< Instruction opcode

    REG_TYPE dest_reg;          ///< Destination register
    WRITE_MASK write_mask;      ///< Destination write mask
    signed char dest_scale;     ///< Destination scale
    bool b_saturate;            ///< \a true if _sat modifier has been used
    bool b_pair;                ///< \a true if it's the second instruction in a pair

    int line_number;            ///< Instruction line number or -1 if line number not available

    SRCREG src_reg[3];          ///< Source registers (max 3)

    unsigned char constant[4];  ///< Constant value for opcode OP_DEF
};

/// Converts a series of tokens read from binary shader file to a series of decoded instructions
/// \param shader_code Binary shader code read from a compiled shader file in form of a series of unsigned 32-bit int tokens
/// \param p_instructions Pointer to an array that will contain a series of decoded instructions on exit
/// \remarks This function throws ShaderError exception if anything goes wrong during decoding
void DisasmShader( const std::vector<unsigned>& shader_code, std::vector<INSTRUCTION>* p_instructions );

/// Converts a sequence of instructions created by DisasmShader() to a readable string
/// \param instructions Decoded instructions
/// \return Human-readable string
std::string GetReadableShader( const std::vector<INSTRUCTION>& instructions );

#endif // __DISASM_SHADER_H_INCLUDED

// vim:ts=4:sts=4:et
