
#include "png_decoder.h"
#include "../Common/tools.h"
#include "libpng-1.2.5/png.h"
#include <cassert>

namespace {

    struct PNG_INPUT {
        const std::vector<char>* p_input;
        size_t i_pos;
    };

    void PNGAPI read_data_fn( png_structp png_ptr, png_bytep* output_buffer, png_size_t size )
    {
        PNG_INPUT* p_struct = reinterpret_cast<PNG_INPUT*>( png_ptr->io_ptr );
        const std::vector<char>& input = *p_struct->p_input;
        size_t& i_pos = p_struct->i_pos;

        if ( i_pos + size > input.size() )
            png_error( png_ptr, "read error" );
        memcpy( output_buffer, &input[i_pos], size );
        i_pos += size;
    }

}

bool PNGDecoder::CanDecode( const std::vector<char>& input )
{
    if ( input.size() < 8 ) return false;
    return ! png_sig_cmp( (png_bytep)&input[0], 0, 8 );
}

void PNGDecoder::DecodeImage( const std::vector<char>& input, std::vector<char>* p_output, int* p_width, int* p_height )
{
    assert( p_output && p_width && p_height );

    png_structp png_ptr = 0;
    png_infop info_ptr = 0;

    try {

        // Create and initialize png struct
        png_ptr = png_create_read_struct( PNG_LIBPNG_VER_STRING, 0, 0, 0 );
        if ( ! png_ptr ) png_throw_error();

        // Create and initialize memory for image information
        info_ptr = png_create_info_struct( png_ptr );
        if ( ! info_ptr ) png_throw_error();

        // Set read function
        PNG_INPUT input_struct = { &input, 0 };
        png_set_read_fn( png_ptr, &input_struct, (png_rw_ptr)read_data_fn );

        // Read entire PNG file
        png_read_png( png_ptr, info_ptr, PNG_TRANSFORM_STRIP_16 | PNG_TRANSFORM_EXPAND, 0 );

        // Get image data
        const size_t width( info_ptr->width );
        const size_t height( info_ptr->height );
        *p_width = int(width), *p_height = int(height);
        p_output->resize( width * height * 4 );
        std::vector<char>::iterator dest_it = p_output->begin();
        assert( info_ptr->bit_depth == 8 );
        switch ( info_ptr->channels ) {

        case 1: // Y
            FOR_INDEX( irow, height ) {
                const char* src = reinterpret_cast<char*>( info_ptr->row_pointers[ irow ] );
                for ( size_t icol=width; icol > 0; icol-- ) {
                    const char gray = *(src++);
                    *(dest_it++) = gray;
                    *(dest_it++) = gray;
                    *(dest_it++) = gray;
                    *(dest_it++) = '\xFF';
                }
            }
            break;

        case 2: // YA
            FOR_INDEX( irow, height ) {
                const char* src = reinterpret_cast<char*>( info_ptr->row_pointers[ irow ] );
                for ( size_t icol=width; icol > 0; icol-- ) {
                    const char gray = *(src++);
                    *(dest_it++) = gray;
                    *(dest_it++) = gray;
                    *(dest_it++) = gray;
                    *(dest_it++) = *(src++);
                }
            }
            break;

        case 3: // RGB
            FOR_INDEX( irow, height ) {
                const char* src = reinterpret_cast<char*>( info_ptr->row_pointers[ irow ] );
                for ( size_t icol=width; icol > 0; icol-- ) {
                    *(dest_it++) = *(src++);
                    *(dest_it++) = *(src++);
                    *(dest_it++) = *(src++);
                    *(dest_it++) = '\xFF';
                }
            }
            break;

        case 4: // RGBA
            FOR_INDEX( irow, height ) {
                const char* src = reinterpret_cast<char*>( info_ptr->row_pointers[ irow ] );
                const size_t row_width = width * 4;
                memcpy( &(*dest_it), src, row_width );
                dest_it += row_width;
            }
            break;

        default:
            assert( 0 );

        }

    } catch ( ... ) {
        png_destroy_read_struct( &png_ptr, info_ptr?(&info_ptr):0, 0 );
        throw;
    }
    if ( png_ptr )
        png_destroy_read_struct( &png_ptr, info_ptr?(&info_ptr):0, 0 );
}

// vim:ts=4:sts=4:et
