
#include "d3d_pixel_shader.h"
#include "../Common/disasm_shader.h"
#include "../Common/tools.h"
#include <map>
#include <cassert>

namespace {

    void SetFunction( std::vector<DWORD>* p_states, DWORD tss, DWORD op, DWORD arg1, DWORD arg2 )
    {
        p_states->push_back( tss );
        p_states->push_back( op );
        p_states->push_back( (tss == D3DTSS_COLOROP) ? D3DTSS_COLORARG1 : D3DTSS_ALPHAARG1 );
        p_states->push_back( arg1 );
        p_states->push_back( (tss == D3DTSS_COLOROP) ? D3DTSS_COLORARG2 : D3DTSS_ALPHAARG2 );
        p_states->push_back( arg2 );
    }

    void SetFunctionRGB( std::vector<DWORD>* p_states, DWORD op, DWORD arg1, DWORD arg2 )
    {
        SetFunction( p_states, D3DTSS_COLOROP, op, arg1, arg2 );
    }

    void SetFunctionA( std::vector<DWORD>* p_states, DWORD op, DWORD arg1, DWORD arg2 )
    {
        SetFunction( p_states, D3DTSS_ALPHAOP, op, arg1, arg2 );
    }

    void SetNop( std::vector<DWORD>* p_states, DWORD tss )
    {
        SetFunction( p_states, tss, D3DTOP_SELECTARG2, D3DTA_TEXTURE, D3DTA_CURRENT );
    }

    void DisableStage( std::vector<DWORD>* p_states, DWORD tss )
    {
        p_states->push_back( STAGESELECT );
        p_states->push_back( tss );
        SetFunctionRGB( p_states, D3DTOP_DISABLE, D3DTA_TEXTURE, D3DTA_CURRENT );
        SetFunctionA( p_states, D3DTOP_DISABLE, D3DTA_TEXTURE, D3DTA_CURRENT );
    }

}

FEATUREMAP::FEATUREMAP( const D3DCAPS9& caps )
{
    max_stages              = caps.MaxTextureBlendStages;
    max_textures            = caps.MaxSimultaneousTextures;
    b_tssargtemp            = (caps.PrimitiveMiscCaps & D3DPMISCCAPS_TSSARGTEMP) != 0;
    b_bumpenvmap            = (caps.TextureOpCaps & D3DTEXOPCAPS_BUMPENVMAP) != 0;
    b_bumpenvmapluminance   = (caps.TextureOpCaps & D3DTEXOPCAPS_BUMPENVMAPLUMINANCE) != 0;
    b_multiplyadd           = (caps.TextureOpCaps & D3DTEXOPCAPS_MULTIPLYADD) != 0;
    b_lerp                  = (caps.TextureOpCaps & D3DTEXOPCAPS_LERP) != 0;
    b_dotproduct3           = (caps.TextureOpCaps & D3DTEXOPCAPS_DOTPRODUCT3) != 0;
    b_blendtexturealphapm   = (caps.TextureOpCaps & D3DTEXOPCAPS_BLENDTEXTUREALPHAPM) != 0;
}

size_t BuildStateList( const std::vector<INSTRUCTION>& instructions, const FEATUREMAP& avail_features, std::vector<DWORD>* p_states, bool b_disable_unused )
{
    assert( ! instructions.empty() && p_states );
    p_states->clear();

    // Index to source instruction array
    size_t src_index = 0;

    // Constant color
    if ( instructions[0].opcode == OP_DEF ) {
        const INSTRUCTION& inst = instructions[0];
        if ( inst.dest_reg != REG_C0 )
            throw ShaderError( "c0 is the only constant register supported", inst.line_number );
        p_states->push_back( D3DRS_TEXTUREFACTOR );
        const unsigned r = inst.constant[0];
        const unsigned g = inst.constant[1];
        const unsigned b = inst.constant[2];
        const unsigned a = inst.constant[3];
        p_states->push_back( (a<<24) | (r<<16) | (g<<8) | b );
        src_index++;
    }

    // Check index
    if ( src_index >= instructions.size() )
        throw ShaderError( "Shader has no instructions", -1 );

    // Extract texture declarations
    std::vector<bool> texture( avail_features.max_stages, false );
    std::vector<unsigned char> bem( avail_features.max_stages, 0 );
    size_t num_textures = 0;
    for ( ; src_index < instructions.size(); src_index++ ) {
        const INSTRUCTION& inst = instructions[src_index];
        const unsigned char opcode = inst.opcode;
        if ( opcode == OP_DEF )
            throw ShaderError( "c0 is the only constant register supported", inst.line_number );
        if ( opcode != OP_TEX && opcode != OP_TEXBEM && opcode != OP_TEXBEML )
            break;
        if ( opcode == OP_TEXBEM && ! avail_features.b_bumpenvmap )
            throw ShaderError( "Bump environment mapping not supported", inst.line_number );
        if ( opcode == OP_TEXBEML && ! avail_features.b_bumpenvmapluminance )
            throw ShaderError( "Bump environment mapping with luminance not supported", inst.line_number );
        const unsigned char stage = inst.dest_reg - REG_T0;
        if ( stage >= avail_features.max_stages )
            throw ShaderError( "Too many shader stages", inst.line_number );
        if ( opcode != OP_TEX && stage == 0 )
            throw ShaderError( "Misplaced environment bump mapping instruction", inst.line_number );
        texture[ stage ] = true;
        num_textures++;
        if ( opcode != OP_TEX && stage > 0 ) {
            const unsigned char src_stage = inst.src_reg[0].reg - REG_T0;
            if ( src_stage != stage-1 )
                throw ShaderError( "Unsupported texbem[l] source", inst.line_number );
            bem[ src_stage ] = opcode;
        }
    }
    if ( num_textures > avail_features.max_textures )
        throw ShaderError( "Too many textures", -1 );

    // Issue arithmetic operations
    size_t stage = 0;
    while ( src_index < instructions.size() ) {
        const INSTRUCTION& inst = instructions[ src_index ];

        // Check stage
        if ( stage >= avail_features.max_stages )
            throw ShaderError( "Too many stages", inst.line_number );

        // Perform bump environment mapping if requested
        if ( bem[stage] ) {
            p_states->push_back( STAGESELECT );
            p_states->push_back( DWORD(stage) );

            DWORD function = (bem[stage] == OP_TEXBEM) ? D3DTOP_BUMPENVMAP : D3DTOP_BUMPENVMAPLUMINANCE;
            SetFunctionRGB( p_states, function, D3DTA_TEXTURE, D3DTA_CURRENT );
            SetNop( p_states, D3DTSS_ALPHAOP );
            p_states->push_back( D3DTSS_RESULTARG );
            p_states->push_back( D3DTA_CURRENT );

            stage++;
            continue;
        }

        // Detect instruction pair
        const bool b_pair_first = (src_index+1 < instructions.size()) && instructions[src_index+1].b_pair;
        const bool b_pair_second = inst.b_pair;
        const bool b_pair = b_pair_first || b_pair_second;

        // Select stage
        if ( ! b_pair_second )
            p_states->push_back( STAGESELECT ),
            p_states->push_back( DWORD(stage) );

        // Output color nop if it's a bare alpha state
        if ( ! b_pair && inst.write_mask == WRITE_A )
            SetNop( p_states, D3DTSS_COLOROP );

        // Extract source operands
        DWORD src_reg[3];
        size_t num_negated = 0;
        size_t num_biased = 0;
        bool b_negate[3] = { false, false, false };
        bool b_bias[3] = { false, false, false };
        FOR_INDEX( i, 3 ) {
            unsigned char reg = inst.src_reg[i].reg;

            // Check if texture register matches stage
            if ( (reg >= REG_T0) && (reg-REG_T0 != stage) )
                throw ShaderError( "Texture register index must correspond to stage index", inst.line_number );

            // Check constant register
            if ( reg == REG_C1 )
                throw ShaderError( "Register c1 not supported", inst.line_number );

            // Check temp register
            if ( reg == REG_R1 && ! avail_features.b_tssargtemp )
                throw ShaderError( "Temporary register not supported", inst.line_number );

            // Convert DIFFUSE to CURRENT for stage 0
            if ( stage == 0 && reg == REG_V0 )
                reg = REG_R0;

            // Determine register type
            switch ( reg ) {
            case REG_R0: reg = D3DTA_CURRENT; break;
            case REG_R1: reg = D3DTA_TEMP; break;
            case REG_V0: reg = D3DTA_DIFFUSE; break;
            case REG_V1: reg = D3DTA_SPECULAR; break;
            case REG_C0: reg = D3DTA_TFACTOR; break;
            default: reg = D3DTA_TEXTURE; break;
            }

            // Check blue replication
            if ( inst.src_reg[i].b_replicate_blue )
                throw ShaderError( "Blue replication swizzle not supported", inst.line_number );

            // Determine alpha replication
            if ( inst.src_reg[i].b_replicate_alpha )
                reg |= D3DTA_ALPHAREPLICATE;

            // Determine complement and extra modifiers
            if ( inst.opcode == OP_DP3 && i < 2 ) {
                switch ( inst.src_reg[i].modifier ) {
                case MOD_NEGATE_BIAS_X2:reg |= D3DTA_COMPLEMENT; break;
                case MOD_BIAS_X2:       break;
                default:
                    throw ShaderError( "dp3 arguments must contain bias", inst.line_number );            
                }
            } else
                switch ( inst.src_reg[i].modifier ) {
                case MOD_NONE:          break;
                case MOD_NEGATE:        b_negate[i] = true; num_negated++; break;
                case MOD_NEGATE_BIAS:   reg |= D3DTA_COMPLEMENT; b_bias[i] = true; num_biased++; break;
                case MOD_BIAS:          b_bias[i] = true; num_biased++; break;
                case MOD_INVERT:        reg |= D3DTA_COMPLEMENT; break;
                default:
                    throw ShaderError( "Modifier not supported", inst.line_number );
                }

            // Save register
            src_reg[i] = reg;
        }

        // Output operation(s)
        const DWORD no_alpha = ~D3DTA_ALPHAREPLICATE;
        switch ( inst.opcode ) {

        case OP_MOV:
            {   if ( inst.dest_scale )
                    throw ShaderError( "Scale not supported in mov instruction", inst.line_number );
                if ( num_negated )
                    throw ShaderError( "Negation not supported in mov instruction", inst.line_number );
                if ( num_biased )
                    throw ShaderError( "Bias not supported in mov instruction", inst.line_number );
                const bool b_texture = (src_reg[0] & D3DTA_SELECTMASK) == D3DTA_TEXTURE;
                const DWORD op = b_texture ? D3DTOP_SELECTARG1 : D3DTOP_SELECTARG2;
                const DWORD arg2 = b_texture ? D3DTA_CURRENT : src_reg[0];
                if ( inst.write_mask != WRITE_A )
                    SetFunctionRGB( p_states, op, D3DTA_TEXTURE, arg2 );
                if ( inst.write_mask != WRITE_RGB )
                    SetFunctionA( p_states, op, D3DTA_TEXTURE, arg2&no_alpha );
            }
            break;

        case OP_ADD:
            {   if ( inst.dest_scale > 1 || inst.dest_scale < 0 )
                    throw ShaderError( "Unsupported scale factor", inst.line_number );
                DWORD op = D3DTOP_ADD;
                if ( inst.dest_scale == 1 ) {
                    if ( num_negated )
                        throw ShaderError( "Negation not supported together with _x2 in add instruction", inst.line_number );
                    if ( num_biased != 1 )
                        throw ShaderError( "Exactly one source operand must have bias together with _x2 in add instruction", inst.line_number );
                    op = D3DTOP_ADDSIGNED2X;
                } else if ( num_negated == 1 ) {
                    if ( num_biased )
                        throw ShaderError( "Bias not supported together with negation in add instruction", inst.line_number );
                    op = D3DTOP_SUBTRACT;
                    if ( b_negate[0] )
                        std::swap( src_reg[0], src_reg[1] );
                } else if ( num_biased == 1 ) {
                    if ( num_negated )
                        throw ShaderError( "Bias not supported together with negation in add instruction", inst.line_number );
                    op = D3DTOP_ADDSIGNED;
                } else {
                    if ( num_negated )
                        throw ShaderError( "Only one source operand can be negated in add instruction", inst.line_number );
                    if ( num_biased )
                        throw ShaderError( "Only one source operand can have bias in add instruction", inst.line_number );
                }
                if ( op != D3DTOP_SUBTRACT && ((src_reg[1] & D3DTA_SELECTMASK) == D3DTA_TEXTURE) )
                    std::swap( src_reg[0], src_reg[1] );
                if ( inst.write_mask != WRITE_A )
                    SetFunctionRGB( p_states, op, src_reg[0], src_reg[1] );
                if ( inst.write_mask != WRITE_RGB )
                    SetFunctionA( p_states, op, src_reg[0]&no_alpha, src_reg[1]&no_alpha );
            }
            break;

        case OP_SUB:
            {   if ( inst.dest_scale )
                    throw ShaderError( "Scale not supported in sub instruction", inst.line_number );
                if ( num_negated )
                    throw ShaderError( "Negation not supported in sub instruction", inst.line_number );
                if ( num_biased )
                    throw ShaderError( "Bias not supported in sub instruction", inst.line_number );
                if ( inst.write_mask != WRITE_A )
                    SetFunctionRGB( p_states, D3DTOP_SUBTRACT, src_reg[0], src_reg[1] );
                if ( inst.write_mask != WRITE_RGB )
                    SetFunctionA( p_states, D3DTOP_SUBTRACT, src_reg[0]&no_alpha, src_reg[1]&no_alpha );
            }
            break;

        case OP_MAD:
            {   if ( inst.dest_scale )
                    throw ShaderError( "Scale not supported in mad instruction", inst.line_number );
                if ( num_negated )
                    throw ShaderError( "Negation not supported in mad instruction", inst.line_number );
                if ( num_biased )
                    throw ShaderError( "Bias not supported in mad instruction", inst.line_number );

                // Detect D3DTOP_ADDSMOOTH
                const DWORD as_arg0 = src_reg[0] & (~D3DTA_COMPLEMENT);
                const DWORD as_arg1 = src_reg[1] & (~D3DTA_COMPLEMENT);
                const DWORD as_arg2 = src_reg[2] & (~D3DTA_COMPLEMENT);
                const bool b_compl0 = (src_reg[0] & D3DTA_COMPLEMENT) != 0;
                const bool b_compl1 = (src_reg[1] & D3DTA_COMPLEMENT) != 0;
                const bool b_compl2 = (src_reg[2] & D3DTA_COMPLEMENT) != 0;
                const bool b_potential_as = (as_arg0 == as_arg2) || (as_arg1 == as_arg2);
                if ( b_potential_as ) {
                    if ( ((as_arg1 == as_arg2) && (b_compl1 != b_compl2)) ||
                        ((as_arg0 == as_arg2) && (b_compl0 != b_compl2)) ) {
                            if ( (as_arg1 == as_arg2) && (b_compl1 != b_compl2) )
                                std::swap( src_reg[0], src_reg[1] );
                            if ( inst.write_mask != WRITE_A )
                                SetFunctionRGB( p_states, D3DTOP_ADDSMOOTH, src_reg[2], src_reg[1] );
                            if ( inst.write_mask != WRITE_RGB )
                                SetFunctionA( p_states, D3DTOP_ADDSMOOTH, src_reg[2]&no_alpha, src_reg[1]&no_alpha );
                            break;
                        }
                }

                // Detect D3DTOP_MODULATE[INV]a_ADDb (a,b - COLOR,ALPHA)
                const DWORD ma_arg0 = src_reg[0] & D3DTA_SELECTMASK;
                const DWORD ma_arg1 = src_reg[1] & D3DTA_SELECTMASK;
                const DWORD ma_arg2 = src_reg[2] & D3DTA_SELECTMASK;
                const bool b_alpha0 = (src_reg[0] & D3DTA_ALPHAREPLICATE) != 0;
                const bool b_alpha1 = (src_reg[1] & D3DTA_ALPHAREPLICATE) != 0;
                const bool b_alpha2 = (src_reg[2] & D3DTA_ALPHAREPLICATE) != 0;
                const bool b_potential_ma = (inst.write_mask == WRITE_RGB) && ((ma_arg0 == ma_arg2) || (ma_arg1 == ma_arg2));
                if ( b_potential_ma ) {
                    const bool b_add_alpha = b_alpha2;
                    if ( b_add_alpha || ((ma_arg0 == ma_arg2) && b_alpha0) || ((ma_arg1 == ma_arg2) && b_alpha1) ) {
                        if ( ! b_add_alpha ) {
                            if ( (ma_arg1 == ma_arg2) && b_alpha1 )
                                std::swap( src_reg[0], src_reg[1] );
                        } else
                            if ( ma_arg1 == ma_arg2 )
                                std::swap( src_reg[0], src_reg[1] );
                        DWORD op;
                        DWORD arg1;
                        const DWORD arg2 = src_reg[1];
                        const bool b_inverse = ((src_reg[0] & D3DTA_COMPLEMENT) != 0) != b_compl2;
                        if ( b_add_alpha ) {
                            op = b_inverse ? D3DTOP_MODULATEINVCOLOR_ADDALPHA : D3DTOP_MODULATECOLOR_ADDALPHA;
                            arg1 = (src_reg[2] & (~D3DTA_ALPHAREPLICATE)) | (src_reg[0] & D3DTA_ALPHAREPLICATE);
                        } else {
                            op = b_inverse ? D3DTOP_MODULATEINVALPHA_ADDCOLOR : D3DTOP_MODULATEALPHA_ADDCOLOR;
                            arg1 = src_reg[2];
                        }
                        SetFunctionRGB( p_states, op, arg1, arg2 );
                        break;
                    }
                }

                // Detect D3DTOP_BLENDTEXTUREALPHAPM
                const bool b_texalpha0 = src_reg[0] == (D3DTA_TEXTURE|D3DTA_ALPHAREPLICATE|D3DTA_COMPLEMENT);
                const bool b_texalpha1 = src_reg[1] == (D3DTA_TEXTURE|D3DTA_ALPHAREPLICATE|D3DTA_COMPLEMENT);
                if ( (b_texalpha0 || b_texalpha1) && avail_features.b_blendtexturealphapm ) {
                    const DWORD arg2 = b_texalpha0 ? src_reg[1] : src_reg[0];
                    if ( inst.write_mask != WRITE_A )
                        SetFunctionRGB( p_states, D3DTOP_BLENDTEXTUREALPHAPM, src_reg[2], arg2 );
                    if ( inst.write_mask != WRITE_RGB )
                        SetFunctionA( p_states, D3DTOP_BLENDTEXTUREALPHAPM, src_reg[2]&no_alpha, arg2&no_alpha );
                }

                // D3DTOP_MULTIPLYADD
                if ( ! avail_features.b_multiplyadd )
                    throw ShaderError( "mad instruction not supported", inst.line_number );
                if ( (src_reg[1] & D3DTA_SELECTMASK) == D3DTA_TEXTURE )
                    std::swap( src_reg[0], src_reg[1] );
                if ( inst.write_mask != WRITE_A )
                    SetFunctionRGB( p_states, D3DTOP_MULTIPLYADD, src_reg[0], src_reg[1] ),
                    p_states->push_back( D3DTSS_COLORARG0 ),
                    p_states->push_back( src_reg[2] );
                if ( inst.write_mask != WRITE_RGB )
                    SetFunctionA( p_states, D3DTOP_MULTIPLYADD, src_reg[0]&no_alpha, src_reg[1]&no_alpha ),
                    p_states->push_back( D3DTSS_ALPHAARG0 ),
                    p_states->push_back( src_reg[2]&no_alpha );
            }
            break;

        case OP_MUL:
            {   if ( num_negated )
                    throw ShaderError( "Negation not supported in mul instruction", inst.line_number );
                if ( num_biased )
                    throw ShaderError( "Bias not supported in mul instruction", inst.line_number );
                if ( inst.dest_scale < 0 )
                    throw ShaderError( "Unsupported scale factor", inst.line_number );
                if ( (src_reg[1] & D3DTA_SELECTMASK) == D3DTA_TEXTURE )
                    std::swap( src_reg[0], src_reg[1] );
                const DWORD op = (inst.dest_scale == 1) ? D3DTOP_MODULATE2X : ((inst.dest_scale == 2) ? D3DTOP_MODULATE4X : D3DTOP_MODULATE);
                if ( inst.write_mask != WRITE_A )
                    SetFunctionRGB( p_states, op, src_reg[0], src_reg[1] );
                if ( inst.write_mask != WRITE_RGB )
                    SetFunctionA( p_states, op, src_reg[0]&no_alpha, src_reg[1]&no_alpha );
            }
            break;

        case OP_DP3:
            {   if ( ! avail_features.b_dotproduct3 )
                    throw ShaderError( "dp3 instruction not supported", inst.line_number );
                if ( inst.write_mask != WRITE_RGBA )
                    throw ShaderError( "dp3 must write all four color channels", inst.line_number );
                // _bx2 is checked when decoding src regs
                if ( (src_reg[1] & D3DTA_SELECTMASK) == D3DTA_TEXTURE )
                    std::swap( src_reg[0], src_reg[1] );
                SetFunctionRGB( p_states, D3DTOP_DOTPRODUCT3, src_reg[0], src_reg[1] );
                SetNop( p_states, D3DTSS_ALPHAOP );
            }
            break;

        case OP_LRP:
            {   if ( inst.dest_scale )
                    throw ShaderError( "Scale not supported in lrp instruction", inst.line_number );
                if ( num_negated )
                    throw ShaderError( "Negation not supported in lrp instruction", inst.line_number );
                if ( num_biased )
                    throw ShaderError( "Bias not supported in lrp instruction", inst.line_number );

                // Try to perform an operation simplier than LERP
                DWORD op = D3DTOP_LERP;
                const DWORD src_reg0 = src_reg[0] & (~D3DTA_COMPLEMENT);
                if ( src_reg0 == (D3DTA_DIFFUSE|D3DTA_ALPHAREPLICATE) )
                    op = D3DTOP_BLENDDIFFUSEALPHA;
                else if ( src_reg0 == (D3DTA_TEXTURE|D3DTA_ALPHAREPLICATE) )
                    op = D3DTOP_BLENDTEXTUREALPHA;
                else if ( src_reg0 == (D3DTA_TFACTOR|D3DTA_ALPHAREPLICATE) )
                    op = D3DTOP_BLENDFACTORALPHA;
                else if ( src_reg0 == (D3DTA_CURRENT|D3DTA_ALPHAREPLICATE) )
                    op = D3DTOP_BLENDCURRENTALPHA;
                if ( op != D3DTOP_LERP && (src_reg[0] & D3DTA_COMPLEMENT) != 0 )
                    std::swap( src_reg[1], src_reg[2] );
                if ( op == D3DTOP_LERP && ! avail_features.b_lerp )
                    throw ShaderError( "lrp instruction not supported", inst.line_number );

                // Output code
                if ( inst.write_mask != WRITE_A ) {
                    SetFunctionRGB( p_states, op, src_reg[1], src_reg[2] );
                    if ( op == D3DTOP_LERP )
                        p_states->push_back( D3DTSS_COLORARG0 ),
                        p_states->push_back( src_reg[0] );
                }
                if ( inst.write_mask != WRITE_RGB ) {
                    SetFunctionA( p_states, op, src_reg[1]&no_alpha, src_reg[2]&no_alpha );
                    if ( op == D3DTOP_LERP )
                        p_states->push_back( D3DTSS_ALPHAARG0 ),
                        p_states->push_back( src_reg[0] );
                }
            }
            break;

        }

        // Output alpha nop if it's a bare color state
        if ( ! b_pair && inst.write_mask == WRITE_RGB )
            SetNop( p_states, D3DTSS_ALPHAOP );

        // Output argument
        if ( ! b_pair_first ) {
            if ( inst.dest_reg == REG_R1 && ! avail_features.b_tssargtemp )
                throw ShaderError( "Temporary register not supported", inst.line_number );
            if ( inst.b_pair && src_index > 0 && instructions[src_index-1].dest_reg != inst.dest_reg )
                throw ShaderError( "Destination register must be the same for both instructions in a pair", inst.line_number );
            p_states->push_back( D3DTSS_RESULTARG );
            p_states->push_back( inst.dest_reg == REG_R0 ? D3DTA_CURRENT : D3DTA_TEMP );
        }

        // Advance to next stage
        if ( ! b_pair_first )
            stage++;

        // Advance to next opcode
        src_index++;
    }
    const size_t num_used = stage;
    if ( b_disable_unused )
        for ( ; stage < avail_features.max_stages; stage++ )
            DisableStage( p_states, DWORD(stage) );
    return num_used;
}

std::string GetReadableTSS( const std::vector<DWORD>& states )
{
    std::map<DWORD,std::string> tss_dest_map;
    tss_dest_map[ D3DTSS_COLOROP ]      = "ColorOp";
    tss_dest_map[ D3DTSS_ALPHAOP ]      = "AlphaOp";
    tss_dest_map[ D3DTSS_COLORARG1 ]    = "ColorArg1";
    tss_dest_map[ D3DTSS_COLORARG2 ]    = "ColorArg2";
    tss_dest_map[ D3DTSS_COLORARG0 ]    = "ColorArg0";
    tss_dest_map[ D3DTSS_ALPHAARG1 ]    = "AlphaArg1";
    tss_dest_map[ D3DTSS_ALPHAARG2 ]    = "AlphaArg2";
    tss_dest_map[ D3DTSS_ALPHAARG0 ]    = "AlphaArg0";
    tss_dest_map[ D3DTSS_RESULTARG ]    = "ResultArg";
    std::map<DWORD,std::string> tss_op_map;
    tss_op_map[ D3DTOP_DISABLE ]        = "Disable";
    tss_op_map[ D3DTOP_SELECTARG1 ]     = "SelectArg1";
    tss_op_map[ D3DTOP_SELECTARG2 ]     = "SelectArg2";
    tss_op_map[ D3DTOP_MODULATE ]       = "Modulate";
    tss_op_map[ D3DTOP_MODULATE2X ]     = "Modulate2x";
    tss_op_map[ D3DTOP_MODULATE4X ]     = "Modulate4x";
    tss_op_map[ D3DTOP_ADD ]            = "Add";
    tss_op_map[ D3DTOP_ADDSIGNED ]      = "AddSigned";
    tss_op_map[ D3DTOP_ADDSIGNED2X ]    = "AddSigned2x";
    tss_op_map[ D3DTOP_SUBTRACT ]       = "Subtract";
    tss_op_map[ D3DTOP_ADDSMOOTH ]      = "AddSmooth";
    tss_op_map[ D3DTOP_DOTPRODUCT3 ]    = "DotProduct3";
    tss_op_map[ D3DTOP_MULTIPLYADD ]    = "MultiplyAdd";
    tss_op_map[ D3DTOP_LERP ]           = "Lerp";
    tss_op_map[ D3DTOP_BLENDDIFFUSEALPHA ]          = "BlendDiffuseAlpha";
    tss_op_map[ D3DTOP_BLENDTEXTUREALPHA ]          = "BlendTextureAlpha";
    tss_op_map[ D3DTOP_BLENDTEXTUREALPHAPM ]        = "BlendTextureAlphaPM";
    tss_op_map[ D3DTOP_BLENDCURRENTALPHA ]          = "BlendCurrentAlpha";
    tss_op_map[ D3DTOP_BLENDFACTORALPHA ]           = "BlendFactorAlpha";
    tss_op_map[ D3DTOP_MODULATEALPHA_ADDCOLOR ]     = "ModulateAlpha_AddColor";
    tss_op_map[ D3DTOP_MODULATECOLOR_ADDALPHA ]     = "ModulateColor_AddAlpha";
    tss_op_map[ D3DTOP_MODULATEINVALPHA_ADDCOLOR ]  = "ModulateInvAlpha_AddColor";
    tss_op_map[ D3DTOP_MODULATEINVCOLOR_ADDALPHA ]  = "ModulateInvColor_AddAlpha";
    tss_op_map[ D3DTOP_BUMPENVMAP ]                 = "BumpEnvMap";
    tss_op_map[ D3DTOP_BUMPENVMAPLUMINANCE ]        = "BumpEnvMapLuminance";
    std::map<DWORD,std::string> tss_arg_map;
    tss_arg_map[ D3DTA_TEXTURE ]        = "Texture";
    tss_arg_map[ D3DTA_DIFFUSE ]        = "Diffuse";
    tss_arg_map[ D3DTA_CURRENT ]        = "Current";
    tss_arg_map[ D3DTA_TFACTOR ]        = "TFactor";
    tss_arg_map[ D3DTA_SPECULAR ]       = "Specular";
    tss_arg_map[ D3DTA_TEMP ]           = "Temp";

    const size_t num = states.size() / 2;
    std::string stage;
    std::string outs;
    FOR_INDEX( i, num ) {
        const DWORD arg = states[2*i];
        const DWORD value = states[2*i+1];
        if ( arg == STAGESELECT ) {
            stage = Int2String( value );
            outs += "\n// Stage " + stage + "\n";
            continue;
        }
        if ( arg == D3DRS_TEXTUREFACTOR ) {
            char s[64]; sprintf( s, "\nTextureFactor = 0x%08X;\n", value );
            outs += s;
            continue;
        }
        outs += tss_dest_map[arg] + "[" + stage + "] = ";
        if ( arg == D3DTSS_COLOROP || arg == D3DTSS_ALPHAOP )
            outs += tss_op_map[value] + ";\n";
        else {
            outs += tss_arg_map[value&D3DTA_SELECTMASK];
            if ( value & D3DTA_ALPHAREPLICATE )
                outs += " | AlphaReplicate";
            if ( value & D3DTA_COMPLEMENT )
                outs += " | Complement";
            outs += ";\n";
        }
    }
    return outs;
}

// vim:ts=4:sts=4:et
