
#ifndef __D3D_PIXEL_SHADER_H_INCLUDED
#define __D3D_PIXEL_SHADER_H_INCLUDED

#define WIN32_LEAN_AND_MEAN
#include <d3d9.h>

#include "../Common/disasm_shader.h"

/// Special stage select constant for BuildStateList()
const unsigned STAGESELECT = 0xFFFFFFFFU;

/// Feature map read from device capabilities
struct FEATUREMAP {
    size_t max_stages;          // D3DCAPS9::MaxTextureBlendStages
    size_t max_textures;        // D3DCAPS9::MaxSimultaneousTextures
    bool b_tssargtemp;          // D3DCAPS9::PrimitiveMiscCaps & D3DPMISCCAPS_TSSARGTEMP
    bool b_bumpenvmap;          // D3DCAPS9::TextureOpCaps & D3DTEXOPCAPS_BUMPENVMAP
    bool b_bumpenvmapluminance; // D3DCAPS9::TextureOpCaps & D3DTEXOPCAPS_BUMPENVMAPLUMINANCE
    bool b_multiplyadd;         // D3DCAPS9::TextureOpCaps & D3DTEXOPCAPS_MULTIPLYADD
    bool b_lerp;                // D3DCAPS9::TextureOpCaps & D3DTEXOPCAPS_LERP
    bool b_dotproduct3;         // D3DCAPS9::TextureOpCaps & D3DTEXOPCAPS_DOTPRODUCT3
    bool b_blendtexturealphapm; // D3DCAPS9::TextureOpCaps & D3DTEXOPCAPS_BLENDTEXTUREALPHAPM

    FEATUREMAP() { }
    FEATUREMAP( const D3DCAPS9& caps );
};

/// Converts a sequence of instructions into D3D TSS settings
/// \param instructions Instruction stream, possibly decoded by DisasmShader()
/// \param avail_features Supported device capabilities
/// \param p_states Pointer to array that will contain TSS settings on exit
/// \param b_disable_unused If \a true, unused stages are disabled
/// \return Returns number of used stages
/// \remarks The function throws ShaderError exception if it cannot find a proper
///          shader conversion to TSS settings. Each entry in the output array
///          consists of two subsequent words: register and value. Possible
///          register values are as follows:
///          <ul>
///          <li>STAGESELECT - selects stage; subsequent values are related to this stage
///          <li>D3DRS_TEXTUREFACTOR - render state - constant (TFACTOR)
///          <li>D3DTSS_COLOROP
///          <li>D3DTSS_ALPHAOP
///          <li>D3DTSS_COLORARGn (n=0,1,2)
///          <li>D3DTSS_ALPHAARGn (n=0,1,2)
///          <li>D3DTSS_RESULTARG
///          </ul>
size_t BuildStateList( const std::vector<INSTRUCTION>& instructions, const FEATUREMAP& avail_features, std::vector<DWORD>* p_states, bool b_disable_unused = true );

/// Converts TSS settings produced by BuildStateList() to a readable string
/// \param states TSS states sequence
/// \return Readable string, containing // comments and \n newlines
/// \remarks The returned string is a valid D3DX effects code
std::string GetReadableTSS( const std::vector<DWORD>& states );

#endif // __D3D_PIXEL_SHADER_H_INCLUDED

// vim:ts=4:sts=4:et
