
#include "gl_pixel_shader.h"
#include "../Common/tools.h"
#include "glext.h"
#include <map>

namespace {

    void SetParam( std::vector<TEXENVVALUE>* p_states, GLenum pname, GLint value )
    {
        TEXENVVALUE param;
        param.param = pname;
        param.valuei = value;
        p_states->push_back( param );
    }

    bool HasNoMods( const SRCREG& src_reg )
    {
        return src_reg.modifier == MOD_NONE && ! src_reg.b_replicate_alpha && ! src_reg.b_replicate_blue;
    }

    void SetFunctionRGB( std::vector<TEXENVVALUE>* p_states, GLint fun, GLint scale, GLint src0, GLint op0, GLint src1=0, GLint op1=0, GLint src2=0, GLint op2=0, GLint src3=0, GLint op3=0 )
    {
        SetParam( p_states, GL_COMBINE_RGB, fun );
        SetParam( p_states, GL_SOURCE0_RGB, src0 );
        SetParam( p_states, GL_OPERAND0_RGB, op0 );
        if ( op1 )
            SetParam( p_states, GL_SOURCE1_RGB, src1 ),
            SetParam( p_states, GL_OPERAND1_RGB, op1 );
        if ( op2 )
            SetParam( p_states, GL_SOURCE2_RGB, src2 ),
            SetParam( p_states, GL_OPERAND2_RGB, op2 );
        if ( op3 )
            SetParam( p_states, GL_SOURCE3_RGB, src3 ),
            SetParam( p_states, GL_OPERAND3_RGB, op3 );
        SetParam( p_states, GL_RGB_SCALE, scale );
    }

    void SetFunctionAlpha( std::vector<TEXENVVALUE>* p_states, GLint fun, GLint scale, GLint src0, GLint op0, GLint src1=0, GLint op1=0, GLint src2=0, GLint op2=0, GLint src3=0, GLint op3=0 )
    {
        SetParam( p_states, GL_COMBINE_ALPHA, fun );
        SetParam( p_states, GL_SOURCE0_ALPHA, src0 );
        SetParam( p_states, GL_OPERAND0_ALPHA, op0 );
        if ( op1 )
            SetParam( p_states, GL_SOURCE1_ALPHA, src1 ),
            SetParam( p_states, GL_OPERAND1_ALPHA, op1 );
        if ( op2 )
            SetParam( p_states, GL_SOURCE2_ALPHA, src2 ),
            SetParam( p_states, GL_OPERAND2_ALPHA, op2 );
        if ( op3 )
            SetParam( p_states, GL_SOURCE3_ALPHA, src3 ),
            SetParam( p_states, GL_OPERAND3_ALPHA, op3 );
        SetParam( p_states, GL_ALPHA_SCALE, scale );
    }

}

GLFEATUREMAP::GLFEATUREMAP()
{
    const char* ext = (const char*)glGetString( GL_EXTENSIONS );
    if ( strstr( ext, "GL_ARB_multitexture" ) ) {
        GLint value = 1;
        glGetIntegerv( GL_MAX_TEXTURE_UNITS_ARB, &value );
        max_texture_units = value;
    } else
        max_texture_units = 1;

    b_texture_env_add = strstr( ext, "GL_ARB_texture_env_add" ) != 0;
    b_texture_env_combine = strstr( ext, "GL_ARB_texture_env_combine" ) != 0;
    b_texture_env_dot3 = strstr( ext, "GL_ARB_texture_env_dot3" ) != 0;
    b_envmap_bumpmap = strstr( ext, "GL_ATI_envmap_bumpmap" ) != 0;
    b_texture_env_combine3 = strstr( ext, "GL_ATI_texture_env_combine3" ) != 0;
    b_texture_env_combine4 = strstr( ext, "GL_NV_texture_env_combine4" ) != 0;
}

size_t BuildTexEnvSetup( const std::vector<INSTRUCTION>& instructions, const GLFEATUREMAP& avail_features, std::vector<TEXENVVALUE>* p_states, bool b_disable_unused )
{
    assert( ! instructions.empty() && p_states );
    p_states->clear();

    // Index to source instruction array
    size_t src_index = 0;

    // Constant colors
    TEXENVVALUE set_color[2];
    bool b_use_const[2] = { false, false };
    set_color[0].param = set_color[1].param = GL_TEXTURE_ENV_COLOR;
    for ( ; src_index < instructions.size() && (instructions[src_index].opcode == OP_DEF); src_index++ ) {
        const INSTRUCTION& inst = instructions[src_index];
        const size_t const_sel = inst.dest_reg - REG_C0;
        if ( b_use_const[const_sel] )
            throw ShaderError( "Constant redefinition", inst.line_number );
        b_use_const[const_sel] = true;
        FOR_INDEX( i, 4 )
            set_color[const_sel].valuef[i] = inst.constant[i] / 255.0f;
    }

    // Check index
    if ( src_index >= instructions.size() )
        throw ShaderError( "Shader has no instructions", -1 );

    // Extract texture declarations
    std::vector<unsigned char> bem( avail_features.max_texture_units, 0 );
    GLint num_textures = 0;
    for ( ; src_index < instructions.size(); src_index++ ) {
        const INSTRUCTION& inst = instructions[src_index];
        const unsigned char opcode = inst.opcode;
        if ( opcode == OP_DEF )
            throw ShaderError( "c0 is the only constant register supported", inst.line_number );
        if ( opcode != OP_TEX && opcode != OP_TEXBEM && opcode != OP_TEXBEML )
            break;
        if ( opcode == OP_TEXBEM && ! avail_features.b_envmap_bumpmap )
            throw ShaderError( "Bump environment mapping not supported", inst.line_number );
        if ( opcode == OP_TEXBEML )
            throw ShaderError( "Bump environment mapping with luminance not supported", inst.line_number );
        const unsigned char stage = inst.dest_reg - REG_T0;
        if ( stage != num_textures )
            throw ShaderError( "Textures must be declared sequentially without gaps", inst.line_number );
        if ( stage >= avail_features.max_texture_units )
            throw ShaderError( "Too many textures", inst.line_number );
        if ( opcode != OP_TEX && stage == 0 )
            throw ShaderError( "Misplaced environment bump mapping instruction", inst.line_number );
        num_textures++;
        if ( opcode != OP_TEX && stage > 0 ) {
            const unsigned char src_stage = inst.src_reg[0].reg - REG_T0;
            if ( src_stage != stage-1 )
                throw ShaderError( "Unsupported texbem source", inst.line_number );
            bem[ src_stage ] = opcode;
        }
    }
    if ( num_textures > avail_features.max_texture_units )
        throw ShaderError( "Too many textures", -1 );

    // Issue arithmetic operations
    int const_sel = -1;
    GLint stage = 0;
    while ( src_index < instructions.size() ) {
        const INSTRUCTION& inst = instructions[ src_index ];

        // Check stage
        if ( stage >= num_textures )
            throw ShaderError( "Too many arithmetic instructions", inst.line_number );

        // Perform bump environment mapping if requested
        if ( bem[stage] ) {
            SetParam( p_states, GL_ACTIVE_TEXTURE, stage );
            SetParam( p_states, GL_TEXTURE_ENV_MODE, GL_COMBINE );

            SetParam( p_states, GL_COMBINE_RGB, GL_BUMP_ENVMAP_ATI );
            SetParam( p_states, GL_BUMP_TARGET_ATI, stage+1 );

            SetFunctionAlpha( p_states, GL_REPLACE, 1, GL_PREVIOUS, GL_SRC_ALPHA );

            stage++;
            continue;
        }

        // Detect instruction pair
        const bool b_pair_first = (src_index+1 < instructions.size()) && instructions[src_index+1].b_pair;
        const bool b_pair_second = inst.b_pair;
        const bool b_pair = b_pair_first || b_pair_second;

        // Select stage
        if ( ! b_pair_second )
            const_sel = -1,
            SetParam( p_states, GL_ACTIVE_TEXTURE, stage );

        // Extract source operands
        GLint src_reg[4];
        GLint src_mod_rgb[4];
        GLint src_mod_alpha[4];
        size_t num_negated = 0;
        size_t num_biased = 0;
        bool b_negate[3] = { false, false, false };
        bool b_bias[3] = { false, false, false };
        bool b_complement[3] = { false, false, false };
        bool b_alpha_replicate[3] = { false, false, false };
        FOR_INDEX( i, 3 ) {
            unsigned char reg = inst.src_reg[i].reg;

            // Check if texture register matches stage
            if ( (reg >= REG_T0) && (reg-REG_T0 != stage) )
                throw ShaderError( "Texture register index must correspond to stage index", inst.line_number );

            // Check temp register
            if ( reg == REG_R1 )
                throw ShaderError( "Register r1 not supported", inst.line_number );

            // Check specular register
            if ( reg == REG_V1 )
                throw ShaderError( "Register v1 not supported", inst.line_number );

            // Check constant
            if ( (reg == REG_C1 && const_sel == 0) || (reg == REG_C0 && const_sel == 1) )
                throw ShaderError( "Only one constant per instruction pair can be used", inst.line_number );

            // Convert GL_PRIMARY_COLOR to GL_PREVIOUS for unit 0
            if ( stage == 0 && reg == REG_V0 )
                reg = REG_R0;

            // Determine register type
            switch ( reg ) {
            case REG_R0: src_reg[i] = GL_PREVIOUS; break;
            case REG_V0: src_reg[i] = GL_PRIMARY_COLOR; break;
            case REG_C0: src_reg[i] = GL_CONSTANT; const_sel = 0; break;
            case REG_C1: src_reg[i] = GL_CONSTANT; const_sel = 1; break;
            default: src_reg[i] = GL_TEXTURE; break;
            }

            // Check blue replication
            if ( inst.src_reg[i].b_replicate_blue )
                throw ShaderError( "Blue replication swizzle not supported", inst.line_number );

            // Determine alpha replication
            if ( inst.src_reg[i].b_replicate_alpha )
                b_alpha_replicate[i] = true;

            // Determine complement and extra modifiers
            if ( inst.opcode == OP_DP3 && i < 2 ) {
                switch ( inst.src_reg[i].modifier ) {
                case MOD_NEGATE_BIAS_X2:b_complement[i] = true; break;
                case MOD_BIAS_X2:       break;
                default:
                    throw ShaderError( "dp3 arguments must contain bias", inst.line_number );            
                }
            } else
                switch ( inst.src_reg[i].modifier ) {
                case MOD_NONE:          break;
                case MOD_NEGATE:        b_negate[i] = true; num_negated++; break;
                case MOD_NEGATE_BIAS:   b_complement[i] = true; b_bias[i] = true; num_biased++; break;
                case MOD_BIAS:          b_bias[i] = true; num_biased++; break;
                case MOD_INVERT:        b_complement[i] = true; break;
                default:
                    throw ShaderError( "Modifier not supported", inst.line_number );
                }
            if ( ! b_complement[i] ) {
                src_mod_rgb[i] = ( ! b_alpha_replicate[i] ) ? GL_SRC_COLOR : GL_SRC_ALPHA;
                src_mod_alpha[i] = GL_SRC_ALPHA;
            } else {
                src_mod_rgb[i] = ( ! b_alpha_replicate[i] ) ? GL_ONE_MINUS_SRC_COLOR : GL_ONE_MINUS_SRC_ALPHA;
                src_mod_alpha[i] = GL_ONE_MINUS_SRC_ALPHA;
            }
        }

        // Check destination register
        if ( inst.dest_reg != REG_R0 )
            throw ShaderError( "Destination register must be r0", inst.line_number );

        // Check scale
        if ( inst.dest_scale > 2 || inst.dest_scale < 0 )
            throw ShaderError( "Unsupported scale factor", inst.line_number );
        const GLint scale = 1 << inst.dest_scale;

        // nVIDIA-specific extension
        const bool b_mad = (inst.opcode == OP_MAD) || (b_pair_first && instructions[src_index+1].opcode == OP_MAD) || (b_pair_second && instructions[src_index-1].opcode == OP_MAD);
        if ( b_mad && ! avail_features.b_texture_env_combine3 && avail_features.b_texture_env_combine4 ) {

            // Set actual command
            if ( ! b_pair_second )
                SetParam( p_states, GL_TEXTURE_ENV_MODE, GL_COMBINE4_NV );

            // Check negation
            if ( num_negated )
                throw ShaderError( "Negation not supported", inst.line_number );

            // Output replace rgb for bare alpha op
            if ( ! b_pair && inst.write_mask == WRITE_A ) 
                SetFunctionRGB( p_states, GL_ADD, scale, GL_PREVIOUS, GL_SRC_COLOR, GL_ZERO, GL_ONE_MINUS_SRC_COLOR, GL_ZERO, GL_SRC_COLOR, GL_ZERO, GL_SRC_COLOR );

            switch ( inst.opcode ) {

            case OP_SUB:
                throw ShaderError( "sub instruction not supported in this combination", inst.line_number );
                break;

            case OP_DP3:
                throw ShaderError( "dp3 instruction not supported in this combination", inst.line_number );
                break;

            case OP_MOV:
                {   const GLint op = num_biased ? GL_ADD_SIGNED : GL_ADD;
                    if ( inst.write_mask != WRITE_A )
                        SetFunctionRGB( p_states, op, scale, src_reg[0], src_mod_rgb[0], GL_ZERO, GL_ONE_MINUS_SRC_COLOR, GL_ZERO, GL_SRC_COLOR, GL_ZERO, GL_SRC_COLOR );
                    if ( inst.write_mask != WRITE_RGB )
                        SetFunctionAlpha( p_states, op, scale, src_reg[0], src_mod_alpha[0], GL_ZERO, GL_ONE_MINUS_SRC_ALPHA, GL_ZERO, GL_SRC_ALPHA, GL_ZERO, GL_SRC_ALPHA );
                }
                break;

            case OP_ADD:
                {   if ( num_biased > 1 )
                        throw ShaderError( "Only one argument to add instruction can have bias", inst.line_number );
                    const GLint op = num_biased ? GL_ADD_SIGNED : GL_ADD;
                    if ( src_reg[1] == GL_TEXTURE )
                        std::swap( src_reg[0], src_reg[1] ),
                        std::swap( src_mod_rgb[0], src_mod_rgb[1] ),
                        std::swap( src_mod_alpha[0], src_mod_alpha[1] );
                    if ( inst.write_mask != WRITE_A )
                        SetFunctionRGB( p_states, op, scale, src_reg[0], src_mod_rgb[0], GL_ZERO, GL_ONE_MINUS_SRC_COLOR, src_reg[1], src_mod_rgb[1], GL_ZERO, GL_ONE_MINUS_SRC_COLOR );
                    if ( inst.write_mask != WRITE_RGB )
                        SetFunctionAlpha( p_states, op, scale, src_reg[0], src_mod_alpha[0], GL_ZERO, GL_ONE_MINUS_SRC_ALPHA, src_reg[1], src_mod_alpha[1], GL_ZERO, GL_ONE_MINUS_SRC_ALPHA );
                }
                break;

            case OP_MUL:
                {   if ( num_biased )
                        throw ShaderError( "Bias not supported in mul instruction", inst.line_number );
                    if ( src_reg[1] == GL_TEXTURE )
                        std::swap( src_reg[0], src_reg[1] ),
                        std::swap( src_mod_rgb[0], src_mod_rgb[1] ),
                        std::swap( src_mod_alpha[0], src_mod_alpha[1] );
                    if ( inst.write_mask != WRITE_A )
                        SetFunctionRGB( p_states, GL_ADD, scale, src_reg[0], src_mod_rgb[0], src_reg[1], src_mod_rgb[1], GL_ZERO, GL_SRC_COLOR, GL_ZERO, GL_SRC_COLOR );
                    if ( inst.write_mask != WRITE_RGB )
                        SetFunctionAlpha( p_states, GL_ADD, scale, src_reg[0], src_mod_alpha[0], src_reg[1], src_mod_alpha[1], GL_ZERO, GL_SRC_ALPHA, GL_ZERO, GL_SRC_ALPHA );
                }
                break;

            case OP_MAD:
                {   if ( num_biased > 1 || (num_biased==1 && ! b_bias[2]) )
                        throw ShaderError( "Bias supported only on last argument to mad instruction", inst.line_number );
                    const GLint op = num_biased ? GL_ADD_SIGNED : GL_ADD;
                    if ( src_reg[1] == GL_TEXTURE )
                        std::swap( src_reg[0], src_reg[1] ),
                        std::swap( src_mod_rgb[0], src_mod_rgb[1] ),
                        std::swap( src_mod_alpha[0], src_mod_alpha[1] );
                    if ( inst.write_mask != WRITE_A )
                        SetFunctionRGB( p_states, op, scale, src_reg[0], src_mod_rgb[0], src_reg[1], src_mod_rgb[1], src_reg[2], src_mod_rgb[2], GL_ZERO, GL_ONE_MINUS_SRC_COLOR );
                    if ( inst.write_mask != WRITE_RGB )
                        SetFunctionAlpha( p_states, op, scale, src_reg[0], src_mod_alpha[0], src_reg[1], src_mod_alpha[1], src_reg[2], src_mod_alpha[2], GL_ZERO, GL_ONE_MINUS_SRC_ALPHA );
                }
                break;

            case OP_LRP:
                {   if ( num_biased )
                        throw ShaderError( "Bias not supported in lrp instruction", inst.line_number );
                    GLint mod_rgb_x;
                    switch ( src_mod_rgb[0] ) {
                    case GL_SRC_COLOR: mod_rgb_x = GL_ONE_MINUS_SRC_COLOR; break;
                    case GL_ONE_MINUS_SRC_COLOR: mod_rgb_x = GL_SRC_COLOR; break;
                    case GL_SRC_ALPHA: mod_rgb_x = GL_ONE_MINUS_SRC_ALPHA; break;
                    case GL_ONE_MINUS_SRC_ALPHA: mod_rgb_x = GL_SRC_ALPHA; break;
                    }
                    GLint mod_alpha_x;
                    switch ( src_mod_alpha[0] ) {
                    case GL_SRC_ALPHA: mod_alpha_x = GL_ONE_MINUS_SRC_ALPHA; break;
                    case GL_ONE_MINUS_SRC_ALPHA: mod_alpha_x = GL_SRC_ALPHA; break;
                    }
                    if ( inst.write_mask != WRITE_A )
                        SetFunctionRGB( p_states, GL_ADD, scale, src_reg[0], src_mod_rgb[0], src_reg[1], src_mod_rgb[1], src_reg[0], mod_rgb_x, src_reg[2], src_mod_rgb[2] );
                    if ( inst.write_mask != WRITE_RGB )
                        SetFunctionAlpha( p_states, GL_ADD, scale, src_reg[0], src_mod_alpha[0], src_reg[1], src_mod_alpha[1], src_reg[0], mod_alpha_x, src_reg[2], src_mod_alpha[2] );
                }
                break;

            }

            // Output replace alpha for bare color op
            if ( ! b_pair && inst.write_mask == WRITE_RGB ) 
                SetFunctionAlpha( p_states, GL_ADD, scale, GL_PREVIOUS, GL_SRC_ALPHA, GL_ZERO, GL_ONE_MINUS_SRC_ALPHA, GL_ZERO, GL_SRC_ALPHA, GL_ZERO, GL_SRC_ALPHA );

            // Output constant for this stage
            if ( ! b_pair_first && const_sel != -1 && b_use_const[const_sel] )
                p_states->push_back( set_color[const_sel] );

            // Advance to next stage
            if ( ! b_pair_first )
                stage++;

            // Advance to next opcode
            src_index++;
            continue;
        }

        // GL_DECAL - special case
        if ( ! b_pair && inst.write_mask == WRITE_RGB && inst.dest_scale == 0 && inst.opcode == OP_LRP ) {
            const SRCREG& src0_rgb = inst.src_reg[0];
            const SRCREG& src1_rgb = inst.src_reg[1];
            const SRCREG& src2_rgb = inst.src_reg[2];
            const REG_TYPE vreg = (stage == 0) ? REG_V0 : REG_R0;
            if ( src0_rgb.reg == REG_T0+stage &&
                 src0_rgb.b_replicate_alpha &&
                 ! src0_rgb.b_replicate_blue &&
                 ((src0_rgb.modifier == MOD_NONE &&
                   src1_rgb.reg == REG_T0+stage &&
                   src2_rgb.reg == vreg) ||
                  (src0_rgb.modifier == MOD_INVERT &&
                   src1_rgb.reg == vreg &&
                   src2_rgb.reg == REG_T0+stage)) &&
                 HasNoMods( src1_rgb ) &&
                 HasNoMods( src2_rgb ) ) {
                SetParam( p_states, GL_TEXTURE_ENV_MODE, GL_DECAL );
                src_index++;
                stage++;
                continue;
            }
        }

        // Original OpenGL texture environment functions
        if ( (b_pair_first || inst.write_mask == WRITE_RGBA) && (inst.dest_scale == 0) ) {
            const INSTRUCTION& inst_rgb = (inst.write_mask != WRITE_A) ? inst : instructions[src_index+1];
            const INSTRUCTION& inst_a = (inst.write_mask != WRITE_RGB) ? inst : instructions[src_index+1];
            const SRCREG& src0_rgb = inst_rgb.src_reg[0];
            const SRCREG& src0_a = inst_a.src_reg[0];
            const SRCREG& src1_rgb = inst_rgb.src_reg[1];
            const SRCREG& src1_a = inst_a.src_reg[1];
            const SRCREG& src2_rgb = inst_rgb.src_reg[2];
            const SRCREG& src2_a = inst_a.src_reg[2];
            const REG_TYPE vreg = (stage == 0) ? REG_V0 : REG_R0;

            // GL_REPLACE
            if ( inst_rgb.opcode == OP_MOV &&
                 src0_rgb.reg == REG_T0+stage &&
                 HasNoMods( src0_rgb ) &&
                 inst_a.opcode == OP_MOV &&
                 src0_a.reg == REG_T0+stage &&
                 HasNoMods( src0_a ) ) {
                SetParam( p_states, GL_TEXTURE_ENV_MODE, GL_REPLACE );
                src_index += b_pair_first ? 2 : 1;
                stage++;
                continue;
            }

            // GL_MODULATE
            if ( inst_rgb.opcode == OP_MUL &&
                 ((src0_rgb.reg == REG_T0+stage && src1_rgb.reg == vreg) ||
                  (src1_rgb.reg == REG_T0+stage && src0_rgb.reg == vreg)) &&
                 HasNoMods( src0_rgb ) &&
                 HasNoMods( src1_rgb ) &&
                 inst_a.opcode == OP_MUL &&
                 ((src0_a.reg == REG_T0+stage && src1_a.reg == vreg) ||
                  (src1_a.reg == REG_T0+stage && src0_a.reg == vreg)) &&
                 HasNoMods( src0_a ) &&
                 HasNoMods( src1_a ) ) {
                SetParam( p_states, GL_TEXTURE_ENV_MODE, GL_MODULATE );
                src_index += b_pair_first ? 2 : 1;
                stage++;
                continue;
            }

            // GL_DECAL
            if ( inst_rgb.opcode == OP_LRP &&
                 src0_rgb.reg == REG_T0+stage &&
                 src0_rgb.b_replicate_alpha &&
                 ! src0_rgb.b_replicate_blue &&
                 ((src0_rgb.modifier == MOD_NONE &&
                   src1_rgb.reg == REG_T0+stage &&
                   src2_rgb.reg == vreg) ||
                  (src0_rgb.modifier == MOD_INVERT &&
                   src1_rgb.reg == vreg &&
                   src2_rgb.reg == REG_T0+stage)) &&
                 HasNoMods( src1_rgb ) &&
                 HasNoMods( src2_rgb ) &&
                 inst_a.opcode == OP_MOV &&
                 src0_a.reg == vreg &&
                 HasNoMods( src0_a ) ) {
                SetParam( p_states, GL_TEXTURE_ENV_MODE, GL_DECAL );
                src_index += b_pair_first ? 2 : 1;
                stage++;
                continue;
            }

            // GL_BLEND
            if ( inst_rgb.opcode == OP_LRP &&
                 src0_rgb.reg == REG_T0+stage &&
                 ! src0_rgb.b_replicate_alpha &&
                 ! src0_rgb.b_replicate_blue &&
                 ((src0_rgb.modifier == MOD_NONE &&
                   (src1_rgb.reg == REG_C0 || src1_rgb.reg == REG_C1) &&
                   src2_rgb.reg == vreg) ||
                  (src0_rgb.modifier == MOD_INVERT &&
                   src1_rgb.reg == vreg &&
                   (src2_rgb.reg == REG_C0 || src2_rgb.reg == REG_C1))) &&
                 HasNoMods( src1_rgb ) &&
                 HasNoMods( src2_rgb ) &&
                 inst_a.opcode == OP_MUL &&
                 ((src0_a.reg == REG_T0+stage && src1_a.reg == vreg) ||
                  (src1_a.reg == REG_T0+stage && src0_a.reg == vreg)) &&
                 HasNoMods( src0_a ) &&
                 HasNoMods( src1_a ) ) {
                if ( const_sel != -1 && b_use_const[const_sel] )
                    p_states->push_back( set_color[const_sel] );
                SetParam( p_states, GL_TEXTURE_ENV_MODE, GL_BLEND );
                src_index += b_pair_first ? 2 : 1;
                stage++;
                continue;
            }

            // GL_ADD
            if ( avail_features.b_texture_env_add &&
                 inst_rgb.opcode == OP_ADD &&
                 ((src0_rgb.reg == REG_T0+stage && src1_rgb.reg == vreg) ||
                  (src1_rgb.reg == REG_T0+stage && src0_rgb.reg == vreg)) &&
                 HasNoMods( src0_rgb ) &&
                 HasNoMods( src1_rgb ) &&
                 inst_a.opcode == OP_MUL &&
                 ((src0_a.reg == REG_T0+stage && src1_a.reg == vreg) ||
                  (src1_a.reg == REG_T0+stage && src0_a.reg == vreg)) &&
                 HasNoMods( src0_a ) &&
                 HasNoMods( src1_a ) ) {
                SetParam( p_states, GL_TEXTURE_ENV_MODE, GL_ADD );
                src_index += b_pair_first ? 2 : 1;
                stage++;
                continue;
            }
        }

        // If not resolved already - use combined functions
        if ( ! avail_features.b_texture_env_combine )
            throw ShaderError( "Feature not available", inst.line_number );
        SetParam( p_states, GL_TEXTURE_ENV_MODE, GL_COMBINE );

        // Output color nop if it's a bare alpha state
        if ( ! b_pair && inst.write_mask == WRITE_A )
            SetFunctionRGB( p_states, GL_REPLACE, 1, GL_PREVIOUS, GL_SRC_COLOR );

        // Output operation(s)
        switch ( inst.opcode ) {

        case OP_MOV:
            {   if ( num_negated )
                    throw ShaderError( "Negation not supported in mov instruction", inst.line_number );
                if ( num_biased )
                    throw ShaderError( "Bias not supported in mov instruction", inst.line_number );
                if ( inst.write_mask != WRITE_A )
                    SetFunctionRGB( p_states, GL_REPLACE, scale, src_reg[0], src_mod_rgb[0] );
                if ( inst.write_mask != WRITE_RGB )
                    SetFunctionAlpha( p_states, GL_REPLACE, scale, src_reg[0], src_mod_alpha[0] );
            }
            break;

        case OP_ADD:
            {   GLint op = GL_ADD;
                if ( num_negated ) {
                    if ( num_biased )
                        throw ShaderError( "Bias not supported together with negation in add instruction", inst.line_number );
                    if ( num_negated != 1 )
                        throw ShaderError( "Only one source operand can be negated in add instruction", inst.line_number );
                    op = GL_SUBTRACT;
                    if ( b_negate[0] )
                        std::swap( src_reg[0], src_reg[1] ),
                        std::swap( src_mod_rgb[0], src_mod_rgb[1] ),
                        std::swap( src_mod_alpha[0], src_mod_alpha[1] );
                } else if ( num_biased ) {
                    if ( num_biased != 1 )
                        throw ShaderError( "Exactly one source operand must have bias together with _x2 in add instruction", inst.line_number );
                    op = GL_ADD_SIGNED;
                }
                if ( op != GL_SUBTRACT && src_reg[1] == GL_TEXTURE )
                    std::swap( src_reg[0], src_reg[1] ),
                    std::swap( src_mod_rgb[0], src_mod_rgb[1] ),
                    std::swap( src_mod_alpha[0], src_mod_alpha[1] );
                if ( inst.write_mask != WRITE_A )
                    SetFunctionRGB( p_states, op, scale, src_reg[0], src_mod_rgb[0], src_reg[1], src_mod_rgb[1] );
                if ( inst.write_mask != WRITE_RGB )
                    SetFunctionAlpha( p_states, op, scale, src_reg[0], src_mod_alpha[0], src_reg[1], src_mod_alpha[1] );
            }
            break;

        case OP_SUB:
            {   if ( num_negated )
                    throw ShaderError( "Negation not supported in sub instruction", inst.line_number );
                if ( num_biased )
                    throw ShaderError( "Bias not supported in sub instruction", inst.line_number );
                if ( inst.write_mask != WRITE_A )
                    SetFunctionRGB( p_states, GL_SUBTRACT, scale, src_reg[0], src_mod_rgb[0], src_reg[1], src_mod_rgb[1] );
                if ( inst.write_mask != WRITE_RGB )
                    SetFunctionAlpha( p_states, GL_SUBTRACT, scale, src_reg[0], src_mod_alpha[0], src_reg[1], src_mod_alpha[1] );
            }
            break;

        case OP_MAD:
            {   if ( ! avail_features.b_texture_env_combine3 )
                    throw ShaderError( "mad instruction not supported", inst.line_number );
                if ( num_negated > 1 || (num_negated == 1 && ! b_negate[2]) )
                    throw ShaderError( "Negation supported only in last argument to mad instruction", inst.line_number );
                if ( num_biased > 1 || (num_biased == 1 && ! b_bias[2]) )
                    throw ShaderError( "Bias supported only in last argument to mad instruction", inst.line_number );

                GLint op = GL_MODULATE_ADD_ATI;
                if ( num_biased )
                    op = GL_MODULATE_SIGNED_ADD_ATI;
                if ( num_negated )
                    op = GL_MODULATE_SUBTRACT_ATI;
                if ( inst.write_mask != WRITE_A )
                    SetFunctionRGB( p_states, op, scale, src_reg[0], src_mod_rgb[0], src_reg[2], src_mod_rgb[2], src_reg[1], src_mod_rgb[1] );
                if ( inst.write_mask != WRITE_RGB )
                    SetFunctionAlpha( p_states, op, scale, src_reg[0], src_mod_alpha[0], src_reg[2], src_mod_alpha[2], src_reg[1], src_mod_alpha[1] );
            }
            break;

        case OP_MUL:
            {   if ( num_negated )
                    throw ShaderError( "Negation not supported in mul instruction", inst.line_number );
                if ( num_biased )
                    throw ShaderError( "Bias not supported in mul instruction", inst.line_number );
                if ( src_reg[1] == GL_TEXTURE )
                    std::swap( src_reg[0], src_reg[1] ),
                    std::swap( src_mod_rgb[0], src_mod_rgb[1] ),
                    std::swap( src_mod_alpha[0], src_mod_alpha[1] );
                if ( inst.write_mask != WRITE_A )
                    SetFunctionRGB( p_states, GL_MODULATE, scale, src_reg[0], src_mod_rgb[0], src_reg[1], src_mod_rgb[1] );
                if ( inst.write_mask != WRITE_RGB )
                    SetFunctionAlpha( p_states, GL_MODULATE, scale, src_reg[0], src_mod_alpha[0], src_reg[1], src_mod_alpha[1] );
            }
            break;

        case OP_DP3:
            {   if ( ! avail_features.b_texture_env_dot3 )
                    throw ShaderError( "dp3 instruction not supported", inst.line_number );
                if ( inst.write_mask == WRITE_A )
                    throw ShaderError( "dp3 cannot go only to alpha channel", inst.line_number );
                if ( src_reg[1] == GL_TEXTURE )
                    std::swap( src_reg[0], src_reg[1] ),
                    std::swap( src_mod_rgb[0], src_mod_rgb[1] ),
                    std::swap( src_mod_alpha[0], src_mod_alpha[1] );
                SetFunctionRGB( p_states, (inst.write_mask == WRITE_RGBA) ? GL_DOT3_RGBA : GL_DOT3_RGB, scale, src_reg[0], src_mod_rgb[0], src_reg[1], src_mod_rgb[1] );
                if ( inst.write_mask == WRITE_RGBA )
                    SetFunctionAlpha( p_states, GL_REPLACE, 1, GL_PREVIOUS, GL_SRC_ALPHA );
            }
            break;

        case OP_LRP:
            {   if ( num_negated )
                    throw ShaderError( "Negation not supported in lrp instruction", inst.line_number );
                if ( num_biased )
                    throw ShaderError( "Bias not supported in lrp instruction", inst.line_number );
                if ( inst.write_mask != WRITE_A )
                    SetFunctionRGB( p_states, GL_INTERPOLATE, scale, src_reg[1], src_mod_rgb[1], src_reg[2], src_mod_rgb[2], src_reg[0], src_mod_rgb[0] );
                if ( inst.write_mask != WRITE_RGB )
                    SetFunctionAlpha( p_states, GL_INTERPOLATE, scale, src_reg[1], src_mod_alpha[1], src_reg[2], src_mod_alpha[2], src_reg[0], src_mod_alpha[0] );
            }
            break;

        }

        // Output alpha nop if it's a bare color state
        if ( ! b_pair && inst.write_mask == WRITE_RGB )
            SetFunctionAlpha( p_states, GL_REPLACE, 1, GL_PREVIOUS, GL_SRC_ALPHA );

        // Output constant for this stage
        if ( ! b_pair_first && const_sel != -1 && b_use_const[const_sel] )
            p_states->push_back( set_color[const_sel] );

        // Advance to next stage
        if ( ! b_pair_first )
            stage++;

        // Advance to next opcode
        src_index++;
    }
    const size_t num_used = stage;
    if ( b_disable_unused )
        for ( ; stage < avail_features.max_texture_units; stage++ ) {
            TEXENVVALUE param;
            param.param = UNITDISABLE;
            param.valuei = stage;
            p_states->push_back( param );
        }
    return num_used;
}

std::string GetReadableTexEnv( const std::vector<TEXENVVALUE>& states )
{
    std::map<GLint,std::string> const_map;
    const_map[ GL_TEXTURE_ENV_COLOR ]   = "TextureEnvColor";
    const_map[ GL_TEXTURE_ENV_MODE ]    = "TextureEnvMode";
    const_map[ GL_COMBINE_RGB ]         = "CombineRGB";
    const_map[ GL_SOURCE0_RGB ]         = "Source0RGB";
    const_map[ GL_OPERAND0_RGB ]        = "Operand0RGB";
    const_map[ GL_SOURCE1_RGB ]         = "Source1RGB";
    const_map[ GL_OPERAND1_RGB ]        = "Operand2RGB";
    const_map[ GL_SOURCE2_RGB ]         = "Source2RGB";
    const_map[ GL_OPERAND2_RGB ]        = "Operand2RGB";
    const_map[ GL_SOURCE3_RGB ]         = "Source3RGB";
    const_map[ GL_OPERAND3_RGB ]        = "Operand3RGB";
    const_map[ GL_RGB_SCALE ]           = "RGBScale";
    const_map[ GL_COMBINE_ALPHA ]       = "CombineAlpha";
    const_map[ GL_SOURCE0_ALPHA ]       = "Source0Alpha";
    const_map[ GL_OPERAND0_ALPHA ]      = "Operand0Alpha";
    const_map[ GL_SOURCE1_ALPHA ]       = "Source1Alpha";
    const_map[ GL_OPERAND1_ALPHA ]      = "Operand2Alpha";
    const_map[ GL_SOURCE2_ALPHA ]       = "Source2Alpha";
    const_map[ GL_OPERAND2_ALPHA ]      = "Operand2Alpha";
    const_map[ GL_SOURCE3_ALPHA ]       = "Source3Alpha";
    const_map[ GL_OPERAND3_ALPHA ]      = "Operand3Alpha";
    const_map[ GL_ALPHA_SCALE ]         = "AlphaScale";
    const_map[ GL_BUMP_TARGET_ATI ]     = "BumpTargetATI";
    const_map[ GL_SRC_COLOR ]           = "SrcColor";
    const_map[ GL_SRC_ALPHA ]           = "SrcAlpha";
    const_map[ GL_ONE_MINUS_SRC_COLOR ] = "OneMinusSrcColor";
    const_map[ GL_ONE_MINUS_SRC_ALPHA ] = "OneMinusSrcAlpha";
    const_map[ GL_REPLACE ]             = "Replace";
    const_map[ GL_MODULATE ]            = "Modulate";
    const_map[ GL_DECAL ]               = "Decal";
    const_map[ GL_BLEND ]               = "Blend";
    const_map[ GL_ADD ]                 = "Add";
    const_map[ GL_COMBINE ]             = "Combine";
    const_map[ GL_COMBINE4_NV ]         = "Combine4NV";
    const_map[ GL_ADD_SIGNED ]          = "AddSigned";
    const_map[ GL_SUBTRACT ]            = "Subtract";
    const_map[ GL_INTERPOLATE ]         = "Interpolate";
    const_map[ GL_DOT3_RGB ]            = "Dot3RGB";
    const_map[ GL_DOT3_RGBA ]           = "Dot3RGBA";
    const_map[ GL_BUMP_ENVMAP_ATI ]     = "BumpEnvmapATI";
    const_map[ GL_MODULATE_ADD_ATI ]    = "ModulateAddATI";
    const_map[ GL_MODULATE_SIGNED_ADD_ATI ] = "ModulateSignedAddATI";
    const_map[ GL_MODULATE_SUBTRACT_ATI ] = "ModulateSubtractATI";
    const_map[ GL_TEXTURE ]             = "Texture";
    const_map[ GL_PREVIOUS ]            = "Previous";
    const_map[ GL_PRIMARY_COLOR ]       = "PrimaryColor";
    const_map[ GL_CONSTANT ]            = "Constant";
    const_map[ GL_ZERO ]                = "Zero";

    std::string stage;
    std::string outs;
    FOR_INDEX( i, states.size() ) {
        const TEXENVVALUE& param = states[i];
        if ( param.param == GL_ACTIVE_TEXTURE ) {
            stage = Int2String( param.valuei );
            outs += "\n// Unit " + stage + "\n";
            continue;
        }
        if ( param.param == UNITDISABLE ) {
            stage = Int2String( param.valuei );
            outs += "\n// Disable unit " + stage + "\n";
            continue;
        }
        outs += const_map[ param.param ] + "[" + stage + "] = ";
        if ( param.param == GL_TEXTURE_ENV_COLOR ) {
            char s[128]; sprintf( s, "{ %G, %G, %G, %G };\n", param.valuef[0], param.valuef[1], param.valuef[2], param.valuef[3] );
            outs += s;
            continue;
        }
        if ( param.param == GL_BUMP_TARGET_ATI || param.param == GL_RGB_SCALE || param.param == GL_ALPHA_SCALE ) {
            outs += Int2String( param.valuei ) + ";\n";
            continue;
        }
        outs += const_map[ param.valuei ] + ";\n";
    }
    return outs;
}

// vim:ts=4:sts=4:et

