
#ifndef __GL_PIXEL_SHADER_H
#define __GL_PIXEL_SHADER_H

#include "../Common/disasm_shader.h"

#define WIN32_LEAN_AND_MEAN     // Exclude rarely-used stuff from Windows headers
#include <winapi.h>
#include <GL/gl.h>

/// Feature map read from extensions string
struct GLFEATUREMAP {
    GLint max_texture_units;        // GL_MAX_TEXTURE_UNITS_ARB (GL_ARB_multitexture)
    bool b_texture_env_add;         // GL_ARB_texture_env_add
    bool b_texture_env_combine;     // GL_ARB_texture_env_combine
    bool b_texture_env_dot3;        // GL_ARB_texture_env_dot3
    bool b_envmap_bumpmap;          // GL_ATI_envmap_bumpmap
    bool b_texture_env_combine3;    // GL_ATI_texture_env_combine3
    bool b_texture_env_combine4;    // GL_NV_texture_env_combine4

    GLFEATUREMAP();
};

/// A single value for TexEnv
struct TEXENVVALUE {
    GLenum param;
    GLint valuei;
    GLfloat valuef[4];
};

/// Special unit disable command for TexEnv
const unsigned UNITDISABLE = 0xFFFFFFFF;

/// Converts a sequence of instructions into GL TexEnv settings
/// \param instructions Instruction stream, possibly decoded by DisasmShader()
/// \param avail_features Supported extensions
/// \param p_states Pointer to array that will contain TexEnv setup on exit
/// \param b_disable_unused If \a true, unused units are disabled
/// \return Returns number of used units
/// \remarks The function throws ShaderError exception if it cannot find a proper
///          shader conversion to TexEnv settings. Each entry in the output array
///          contains a setup value for TexEnv. Possible values are as follows:
///          <ul>
///          <li>ACTIVETEXTURE      - selects unit
///          <li>UNITDISABLE        - disables unit
///          <li>TEXTURE_ENV_MODE   - basic instruction
///          <li>COMBINE_RGB        - function on the color pipeline
///          <li>COMBINE_ALPHA      - function on the alpha pipeline
///          <li>SOURCEn_RGB        - source register
///          <li>SOURCEn_ALPHA      - source register
///          <li>OPERANDn_RGB       - source modifier
///          <li>OPERANDn_ALPHA     - source modifier
///          <li>RGB_SCALE          - result scale
///          <li>ALPHA_SCALE        - result scale
///          <li>TEXTURE_ENV_COLOR  - [4 floats] constant color
///          </ul>
size_t BuildTexEnvSetup( const std::vector<INSTRUCTION>& instructions, const GLFEATUREMAP& avail_features, std::vector<TEXENVVALUE>* p_states, bool b_disable_unused = true );

/// Converts GL TexEnv settings produced by BuildTexEnvSetup() to a readable string
/// \param states TSS states sequence
/// \return Readable string, containing // comments and \n newlines
std::string GetReadableTexEnv( const std::vector<TEXENVVALUE>& states );

#endif // __GL_PIXEL_SHADER_H

// vim:ts=4:sts=4:et
